#############################################################################
# stringreq.tcl Copyright 1998, 2004, 2005, 2006, 2007 r0lZ
# String requester.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

# stringReq 1.1 - String requester, by r0lZ, September 1998 - February 2004
#
# Usage: stringreq -window .path args...
#
# Returns the entered string, or "" if user cancelled the requester or the window
# were destroyed.  When used without -window argument, puts string and exit Wish.
#
# Arguments:
# -callback procname Procedure to be called to verify the validity of user
#				input.  The procedure is called with the current content
#				of the string every time the user modify it, and must
#				return true to accept or false to reject the string.
# -cancel str	Text for the Cancel button.
# -def str		Default string in entry.
# -font str		Font name (current system font is the default)
# -geometry str	Geometry descriptor string (in the form XxY+W+H), or
#				"s" to center the requester on screen (the default), or
#				"w" to put it in the topleft corner of the parent window, or
#				"c" to center the requester under the mouse cursor.
# -hidechar char Character used to hide what the user types in.
# -highlight	Flag: highlight text
# -justify opt	"left" (the default), "center" or "right"
# -ok str		Text for the OK button.
# -text str		Message to display in dialog.
# -title str	Title to display in dialog's decorative frame.
# -width nchars	Width of text box in characters or 0 for auto-resize.
#				Default: 32
# -window		Window path to use for dialog top-level.
#				If omitted, this is the root window that is used, and
#				destroyed at end!  Use this for DOS batch inputs.
# -parent		Parent window.  This window is forced on top when stringreq returns.
# -cancelreturn Text to return if the user cancelled the requester (default: "").

# TODO:
#	- Catch all possible errors.
#	- Return an error code 1 when req is destroyed. (?)

namespace eval ::stringreq {
namespace export stringreq

####### main proc
proc stringreq { args } {
	variable ::stringreq::textvar ::stringreq::done

	### args
	set _window ""
	set _parent ""
	set _title ""
	set _text ""
	set _def ""
	set _callback ""
	set _width 32
	set _justify "left"
	set _hidechar ""
	set _ok "OK"
	set _cancel "Cancel"
	set _geometry "s"
	set _highlight ""
	set _font ""
	set _cancelreturn ""
	set _highlight 0

	set ind 0
	set max [llength $args]
	while {$ind < $max} {
		switch -exact -- [lindex $args $ind] {
		  "-window"		{incr ind;set _window	[lindex $args $ind];incr ind}
		  "-parent"		{incr ind;set _parent	[lindex $args $ind];incr ind}
		  "-title"		{incr ind;set _title	[lindex $args $ind];incr ind}
		  "-text"		{incr ind;set _text		[lindex $args $ind];incr ind}
		  "-def"		{incr ind;set _def		[lindex $args $ind];incr ind}
		  "-callback"	{incr ind;set _callback	[lindex $args $ind];incr ind}
		  "-width"		{incr ind;set _width	[lindex $args $ind];incr ind}
		  "-justify"	{incr ind;set _justify	[lindex $args $ind];incr ind}
		  "-hidechar"	{incr ind;set _hidechar	[lindex $args $ind];incr ind}
		  "-ok"			{incr ind;set _ok		[lindex $args $ind];incr ind}
		  "-cancel"		{incr ind;set _cancel	[lindex $args $ind];incr ind}
		  "-geometry"	{incr ind;set _geometry	[lindex $args $ind];incr ind}
		  "-font"		{incr ind;set _font		[lindex $args $ind];incr ind}
		  "-cancelreturn" {incr ind;set _cancelreturn [lindex $args $ind];incr ind}
		  "-highlight"	{incr ind;set _highlight 1}
		  default	{show_warning -icon error -title "Tcl/Tk error" -parent . \
			  -message "stringReq proc:\nunknown option [lindex $args $ind]"; \
			  return ""}
		}
	}

	if {$_window == ""} {
		set w "."
		set wp ""
	} else {
		set w $_window
		set wp $_window
	}

	### init
	set ::stringreq::done -1
	set ::stringreq::textvar $_def

	### build window
	if {$w != "."} {
		catch {destroy $w}
		toplevel $w -class Dialog
	}
	wm withdraw $w
	wm title $w $_title
	wm protocol $w WM_DELETE_WINDOW {set ::stringreq::done 0}
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	if {$w != "."} {
		if {$::config(tk) != "aqua"} {wm transient $w .}
	}
	bind $w <Escape> "set ::stringreq::done 0"
	bind $w <Return> "set ::stringreq::done 1"

	### build widgets

	# row 1: label
	if {$_text != ""} {
		if {!$_highlight} {
			label $wp.label -takefocus 0 -text $_text -padx 1m -justify $_justify
		} else {
			label $wp.label -takefocus 0 -text $_text -padx 1m -pady 2m \
				-foreground brown -background $::config(system_window_color) -justify $_justify \
				-borderwidth 2 -relief groove
		}
		pack $wp.label -expand true -fill both -padx 2 -pady 2
	}

	# row 2: entry
	entry $wp.entry -textvariable ::stringreq::textvar -width $_width \
		-show $_hidechar -justify $_justify
	pack $wp.entry -fill x  -padx 4 -pady 2

	# row 3: buttons
	frame $wp.bf;
	  set width [string length $_ok]
	  if {[string length $_cancel] > $width} {set width [string length $_cancel]}
	  # col 1: resizable blank space
	  pack [frame  $wp.bf.pad1] -side left -fill both -expand true
	  # col 2: OK button
	  button $wp.bf.okbutton -text $_ok -width $width -command {set ::stringreq::done 1}
	  pack $wp.bf.okbutton -side left -fill both -expand true
	  # col 3: Cancel button
	  button $wp.bf.cancelbutton -text $_cancel -width $width -command {set ::stringreq::done 0}
	  pack $wp.bf.cancelbutton -side left -fill both -expand true -padx 4
	  if {$::config(platform) == "mac"} {
		  pack [frame $wp.bf.macpad] -side left -padx 6
	  }

	pack $wp.bf -fill both -pady 4

	### Cosmetics stuff: set selection and focus to string gadget,
	### change window position if requested, and deiconify window
	#wm resizable $w yes no
	if {$_font != ""} {
		if {$_text != ""} {
			$wp.label configure -font $_font
		}
		$wp.entry configure -font $_font
		$wp.bf.okbutton configure -font $_font
		$wp.bf.cancelbutton configure -font $_font
	}

	update
	if {$w != "."} {
		switch $_geometry {
			"" {
				break
			}
			"c" {
				set x [expr {[winfo pointerx .]-([winfo reqwidth $w]/2)}]
				set y [expr {[winfo pointery .]-([winfo reqheight $w]/2)-20}]
				if {$x < 5} {set x 5}
				if {$y < 5} {set y 5}
				wm geom $w +$x+$y
			}
			"w" {
				wm geom $w +[winfo rootx .]+[winfo rooty .]
			}
			"s" {
				set x [expr {{[winfo screenwidth $w]/2 \
					- [winfo reqwidth $w]/2 \
					- [winfo vrootx [winfo parent $w]]}}]
				set y [expr {{[winfo screenheight $w]/2 \
					- [winfo reqheight $w]/2 \
					- [winfo vrooty [winfo parent $w]]}}]
				if {$x < 5} {set x 5}
				if {$y < 5} {set y 5}
				wm geom $w +$x+$y
			}
			default {
				catch [wm geom $w $_geometry]
			}
		}
	} else {
		switch $_geometry {
			"" {
				break
			}
			"c" {
				set x [expr {[winfo pointerx .]-([winfo reqwidth $w]/2)}]
				set y [expr {[winfo pointery .]-([winfo reqheight $w]/2)-20}]
				if {$x < 5} {set x 5}
				if {$y < 5} {set y 5}
				wm geom $w +$x+$y
			}
			"w" {
				break
			}
			"s" {
				set x [expr {{[winfo screenwidth $w]/2 \
					- [winfo reqwidth $w]/2 \
					- [winfo vrootx $w]}}]
				set y [expr {{[winfo screenheight $w]/2 \
					- [winfo reqheight $w]/2 \
					- [winfo vrooty $w]}}]
				if {$x < 5} {set x 5}
				if {$y < 5} {set y 5}
				wm geom $w +$x+$y
			}
			default {
				catch [wm geom $w $_geometry]
			}
		}
	}
	$wp.entry selection range 0 [string length $::stringreq::textvar]
	$wp.entry icursor [string length $::stringreq::textvar]
	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [expr {[winfo reqwidth $w]*4}] [winfo reqheight $w]
	if {$w != "."} {
		if {$::config(tk) != "aqua" && $w != "."} {wm transient $w .}
	}
	bind $wp.entry <Control-z> "set ::stringreq::textvar $_def"
	wm deiconify $w
	if {$w != "."} {
		grab set $w
	}
	focus $w
	focus $wp.entry

	### establist a callback on the variable that holds the string
	### to be able to correct it when a mistake is typed
	if {$_callback != ""} {
		$wp.entry configure -validatecommand "$_callback %P" -validate key
	} else {
		$wp.entry configure -validatecommand { } -validate none
	}

	### wait user
	tkwait variable ::stringreq::done

	### cleanup
	if {$::stringreq::done} {
		set ret $::stringreq::textvar
	} else {
		set ret $_cancelreturn
	}
	unset ::stringreq::textvar
	if {$w != "."} {
		unset ::stringreq::done
		grab release $w
		destroy $w
		if {$_parent != ""} {
			wm deiconify $_parent
			focus $_parent
		}
	} else {
		puts $ret
		exit $::stringreq::done
	}
	tolog "stringreq $_title -> $ret"
	return $ret
}

}	;# end of namespace ::stringreq




##############################################################################
# MSGBOX - Alternate version of tk_messageBox - almost same syntax and usage.
# Additional options:
# -window $w ;# specify the toplevel pathname (default: .msgbox)
# -skip "ID string" ;# Adds a "Do not display this dialog any more" checkbox.
#  If the user ticks the checkbox, the dialog identified by its ID string is
#  added to the skip list ::config(skip_dialogs) and not diasplayed the next
#  time.  The default button is returned when the dialog is hidden.
# -type type ;# in addition to the predefined types of tk_messageBox, it is
#  also possible to specify a list of buttons labels separated by |, such as
#  "Yes|No|No to All".  The returned string is the button label converted to
#  lower case.
# -icon icon ;# In addition to info, question, warning and error, it is now
#  possible to specify any predefined Tk bitmap, such as "hourglass".
##############################################################################

namespace eval ::msgbox {

# example:
# ::msgbox::MsgBox -message test -type yesnocancel
proc MsgBox { args } {
	### args
	set _window .msgbox
	set _parent .
	set _icon info
	set _title PgcEdit
	set _message ""
	set _type ok
	set _default ""
	set _skip ""

	set defaultbuttonnames {yes ok no cancel abort retry ignore}

	#########################################################################
	# @translators: translate the default button names in your language here:
	# (DO NOT translate the string in lower case above!)
	set translatedbuttonnames {Yes OK No Cancel Abort Retry Ignore}

	set ind 0
	set max [llength $args]
	while {$ind < $max} {
		switch -exact -- [lindex $args $ind] {
		  "-window"		{incr ind;set _window	[lindex $args $ind];incr ind; set newargs true}
		  "-parent"		{incr ind;set _parent	[lindex $args $ind];incr ind}
		  "-icon"		{incr ind;set _icon		[lindex $args $ind];incr ind}
		  "-title"		{incr ind;set _title	[lindex $args $ind];incr ind}
		  "-message"	{incr ind;set _message	[lindex $args $ind];incr ind}
		  "-type"		{incr ind;set _type		[lindex $args $ind];incr ind}
		  "-default"	{incr ind;set _default	[string tolower [lindex $args $ind]];incr ind}
		  "-skip"		{incr ind;set _skip		[lindex $args $ind];incr ind; set newargs true}
		  default	{show_warning -icon error -title "Tcl/Tk error" -parent . \
			  -message "MsgBox proc:\nunknown option [lindex $args $ind]"; \
			  return ""}
		}
	}
	if {! [info exists ::config(skip_dialogs)]} {
		set ::config(skip_dialogs) {}
	}
	if {! $::config(show_hidden_dialogs) && $_skip != "" && [lsearch $::config(skip_dialogs) $_skip] != -1} {
		tolog "MsgBox \"$_skip\" skipped: $_default"
		return $_default
	}

	if {[string first "PgcEdit" $_title] == -1 && [string length $_title] < 30} {
		set _title "PgcEdit: $_title"
	}

	if {! $::config(use_internal_dialogs)} {
		if {[lsearch {abortretryignore ok okcancel retrycancel yesno yesnocancel} $_type] == -1} {
			set newargs true
		}
		if {[lsearch {error info question warning} $_icon] == -1} {
			set newargs true
		}
		if {! $newargs} {
			return [tk_messageBox -parent $_parent -icon $_icon -title $_title -message $_message -type $_type -default $_default]
		}
	}

	if {[string first "|" $_type] >= 0} {
		foreach i [split $_type "|"] {
			lappend buttons [list [string tolower $i] $i]
		}
	} else {
		set buttons {}
		set idx 0
		foreach i $defaultbuttonnames {
			if {[string first $i $_type] >= 0} {
				lappend buttons [list [lindex $defaultbuttonnames $idx] [lindex $translatedbuttonnames $idx]]
			}
			incr idx
		}
	}
	if {[llength $buttons] == 0} {
		set buttons [list [string tolower $_type] $_type]
	}
	set maxwidth 8
	foreach i $buttons {
		if {[string length [lindex $i 1]] > $maxwidth} {
			set maxwidth [string length [lindex $i 1]]
		}
	}

	for {set i 0} {$i < 1000} {incr i} {
		set done_control_var done$i
		if {![info exists ::msgbox::$done_control_var]} break
	}
	set ::msgbox::$done_control_var $_default

	### play sound
	if {$::config(internal_dialogs_sounds)} {
		after 50 "::msgbox::play_audio_for_dialog $_icon"
	}

	### build window
	set w $_window
	set i 1
	while {[winfo exists $w]} {
		set w $_window$i
		incr i
	}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $_title
	wm protocol $w WM_DELETE_WINDOW "set ::msgbox::$done_control_var return_false"
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	if {$w != "."} {
		if {$::config(tk) != "aqua"} {wm transient $w .}
	}
	bind $w <Return> "set ::msgbox::$done_control_var return_true"
	bind $w <Escape> "set ::msgbox::$done_control_var return_false"

	### build widgets
	switch $_icon {
		"error" {set color red}
		"warning" {set color orange}
		"question" {set color lightBlue}
		"info" {set color lightGreen}
		default {set color gray90}
	}

	pack [frame $w.top -bd 2 -relief groove] -side top -anchor w -padx 2 -pady 2 -fill x
	pack [label $w.top.icon -bitmap $_icon -bd 2 -relief ridge -width 31 -height 31 -bg $color] -side left -padx 4 -pady 6 -anchor n
	pack [label $w.top.msg -text $_message -justify left -text $_message -wraplength 0] -side left -padx 2 -pady 4 -fill x -expand 1
	pack [frame $w.pad] -side top -anchor w -fill both -expand 1
	if {$_skip != ""} {
		pack [checkbutton $w.skip -text "Do not show this dialog any more." -variable ::msgbox::skip] -anchor w
		if {[lsearch $::config(skip_dialogs) $_skip] != -1} {
			set ::msgbox::skip 1
		} else {
			set ::msgbox::skip 0
		}
	}
	menu $w.menu -tearoff false
	$w.menu add command -label "Copy this message to the clipboard" -command "::msgbox::copy_to_clip $w $_icon [list $buttons]"
	bind $w $::config(popups_button) "tk_popup $w.menu \[winfo pointerx $w\] \[winfo pointery $w\]"

	pack [frame $w.buts] -side top -anchor e -padx 2 -pady 2
	update
	set winwidth [winfo reqwidth $w]
	if {$winwidth < 250} {set winwidth 250}
	set idx 0
	set defaultwidget ""
	foreach i $buttons {
		pack [button $w.buts.but$idx -text [lindex $i 1] \
				-command "set ::msgbox::$done_control_var \"[lindex $i 0]\"" -width $maxwidth] -side left -padx 2 -fill y
		if {$_default == [lindex $i 0]} {
			set defaultwidget $w.buts.but$idx
			set return_true_string [lindex $i 0]
			$w.buts.but$idx configure -borderwidth 3
		}
		set cancelwidget $w.buts.but$idx
		incr idx
	}
	if {$defaultwidget == ""} {
		set defaultwidget $w.buts.but0
		set return_true_string [lindex [lindex $buttons 0] 0]
		$w.buts.but0 configure -borderwidth 3
	}
	update
	if {[winfo reqwidth $w] > $winwidth} {
		set idx 0
		foreach i $buttons {
			$w.buts.but$idx configure -width 0
			incr idx
		}
	}
	if {$_skip != "" && [llength $buttons]>1} {
		update
		set ww [winfo reqwidth $w]
		if {$ww < 250} {set ww 250}
		$w.skip configure -text "Do not show this dialog any more.  Answer \"[$defaultwidget cget -text]\" each time."
		update
		if {[winfo reqwidth $w] > $ww} {
			$w.skip configure -text "Do not show this dialog any more.\nAnswer \"[$defaultwidget cget -text]\" each time." -justify left
		}
	}

	wm minsize $w 250 100
	#wm maxsize $w 750 550
	update
	set reqwidth [winfo reqwidth $w]
	if {$reqwidth < 250} {set reqwidth 250}
	set reqheight [winfo reqheight $w]
	if {$reqheight < 100} {set reqheight 100}
	if {$::config(standard_dialogs_under_mouse)} {
		wm deiconify $w
		update
		set posx [expr [winfo rootx $w]-[winfo x $w]+[winfo x $w.buts]+[winfo x $defaultwidget]+([winfo width $defaultwidget]/2)]
		set posy [expr [winfo rooty $w]-[winfo y $w]+[winfo height $w]-([winfo height $defaultwidget]/2)]
		set wx [expr {[winfo pointerx $w]-$posx}]
		set wy [expr {[winfo pointery $w]-$posy}]
	} else {
		set wx [expr {[winfo rootx $_parent]+(([winfo width  $_parent]-$reqwidth)/2)}]
		set wy [expr {[winfo rooty $_parent]+(([winfo height $_parent]-$reqheight)/2)-25}]
	}
	if {$wx < 10} {set wx 10}
	if {[expr $wx+$reqwidth+10] > [winfo screenwidth $w]} {
		set wx [expr [winfo screenwidth $w]-$reqwidth-10]
		if {$wx < 0} {set wx 0}
	}
	if {$wy < 22} {set wy 22}
	if {[expr $wy+$reqheight+80] > [winfo screenheight $w]} {
		set wy [expr [winfo screenheight $w]-$reqheight-80]
		if {$wy < 0} {set wy 0}
	}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	if {$_parent != ""} {
		grab set $w
	}
	focus $defaultwidget

	### wait user
	tkwait variable ::msgbox::$done_control_var

	if {$_skip != ""} {
		set skipidx [lsearch $::config(skip_dialogs) $_skip]
		if {$skipidx == -1} {
			if {$::msgbox::skip} {
				lappend ::config(skip_dialogs) $_skip
				set ::config(skip_dialogs) [lsort -dictionary $::config(skip_dialogs)]
			}
		} else {
			if {! $::msgbox::skip} {
				set ::config(skip_dialogs) [lreplace $::config(skip_dialogs) $skipidx $skipidx]
			}
		}
	}

	set done [subst \$::msgbox::$done_control_var]
	if {$done == "return_true"} {
		$defaultwidget configure -relief sunken
		after 300 {set ::pause true}
		tkwait variable ::pause
		set ret $return_true_string
	} elseif {$done == "return_false"} {
		$cancelwidget configure -relief sunken
		after 300 {set ::pause true}
		tkwait variable ::pause
		set ret [lindex [lindex $buttons end] 0]
	} else {
		set ret $done
	}

	### cleanup
	unset ::msgbox::$done_control_var
	destroy $w
	if {$_parent != ""} {
		wm deiconify $_parent
		focus $_parent
	}
	tolog "MsgBox \"$_title\" -> $ret"
	return $ret
}

proc play_audio_for_dialog {icontype} {
	if {$::tcl_platform(platform) != "windows" || [catch {package require snack}] || \
			[catch {package require registry}]} {
		bell
		return
	}
	if {[info commands ::snacksounds($icontype)] == ""} {
		switch $icontype {
			"error" {set soundkey SystemHand}
			"warning" {set soundkey SystemExclamation}
			"question" {set soundkey SystemQuestion}
			"info" {set soundkey SystemAsterisk}
			default {bell; return}
		}
		set key "HKEY_CURRENT_USER\\AppEvents\\Schemes\\Apps\\.Default\\$soundkey\\.Current"
		if {[catch {set soundfile [registry get $key ""]}]} {
			bell; return
		}
		set soundfilename [file tail $soundfile]
		if {[string toupper [file extension $soundfile]] != ".WAV"} {
			bell; return
		}
		if {! [file exists $soundfile]} {
			# substiture environment variables in filename
			set idx1 [string first % $soundfile]
			while {$idx1 > -1} {
				set idx2 [string first % $soundfile [expr $idx1+1]]
				if {$idx2 > -1} {
					set str ""
					catch {set str $::env([string range $soundfile [expr $idx1+1] [expr $idx2-1]])}
					set s ""
					if {$idx1 > 0} {set s [string range $soundfile 0 [expr $idx1-1]]}
					set s $s$str
					if {$idx2 < [expr [string length $soundfile]-1]} {set s $s[string range $soundfile [expr $idx2+1] end]}
					if {[file exists $s]} {set soundfile $s}
					set idx1 [string first % $soundfile]
				} else {
					break
				}
			}
		}
		if {! [file exists $soundfile] && [info exists ::env(SystemRoot)]} {
			set soundfile [file join [file join $::env(SystemRoot) "Media"] $soundfilename]
		}
		if {! [file exists $soundfile] && [info exists ::env(windir)]} {
			set soundfile [file join [file join $::env(windir) "Media"] $soundfilename]
		}
		if {! [file exists $soundfile] && [info exists ::env(HOMEDRIVE)]} {
			set soundfile [file join [file join $::env(HOMEDRIVE) "Windows/Media"] $soundfilename]
		}
		if {! [file exists $soundfile] && [info exists ::env(SystemDrive)]} {
			set soundfile [file join [file join $::env(SystemDrive) "Windows/Media"] $soundfilename]
		}
		if {! [file exists $soundfile]} {
			set soundfile [file join "C:/Windows/Media" $soundfilename]
		}
		if {! [file exists $soundfile]} {
			set mediapath ""
			catch {set mediapath [registry get "HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion" "MediaPath"]}
			set soundfile [file join $mediapath $soundfilename]
		}
		if {[file exists $soundfile]} {
			catch {::snack::sound ::snacksounds($icontype) -load $soundfile}
		}
	}
	if {[info commands ::snacksounds($icontype)] != ""} {
		::snacksounds($icontype) play
	} else {
		bell
	}
}

proc copy_to_clip {w icontype buttons} {
	clipboard clear
	clipboard append "Dialog title: \"[wm title $w]\"\nDialog type: $icontype\nDialog message:\n"
	clipboard append [$w.top.msg cget -text]\n
	set txt "Dialog buttons:"
	foreach i $buttons {
		set txt "$txt \[[lindex $i 1]\]"
	}
	clipboard append $txt\n
}

##############################################################################
# Replace the original tk_messageBox function by MsgBox
##############################################################################

rename ::tk_messageBox ::orig_tk_messageBox
proc ::tk_messageBox args {
	uplevel 1 ::msgbox::MsgBox $args
}

}	;# end of namespace ::msgbox

